#include <iostream>
#include <omp.h>
#include <fstream>
#include <bitset>
#include <cmath>

constexpr int nbits=3072;
constexpr int file_size = 30730000;
constexpr int count_of_data=500;
using namespace std;

double get_mean(double** image,int count_of_image)
{
    int sum{0};
    int i{0}, j{0};
#pragma omp parallel shared(image,count_of_image,sum) private(i, j)
    {

#pragma omp for reduction(+:sum)
        for (i = 0; i < count_of_image; ++i) {
            for (j = 0; j < nbits; ++j) {
                sum += static_cast<int>(image[i][j]) ;
            }
        }
    }

    return static_cast<double>(sum)/ static_cast<double>(nbits*count_of_image);
}

double get_std(double mean, int count_of_im, double ** image)
{
    double st=0;
    for (int i = 0; i < count_of_im; ++i) {
        for (int j = 0; j < nbits; ++j) {
            st+=(image[i][j]-mean)*(image[i][j]-mean);
        }
    }
    return sqrt(st);
}

void get_images(string name, int count_of_im, double ** category, double** image, int num_threads)
{

    cout<<__FUNCTION__ <<": Loading..."<<endl;
    ifstream *file = new ifstream(name.c_str(), ios_base::binary | ios_base::in);

    char * buffer = new char[file_size];

    file->seekg(0, std::ios::beg);
    file->read(buffer,file_size);
    file->close();
    cout<<__FUNCTION__ <<": File size: "<<file_size<<endl;

    int i=0, j=0;



    cout<<__FUNCTION__ <<": Preaparing..."<<endl;
    omp_set_dynamic(0);
    omp_set_num_threads(num_threads);
#pragma omp parallel shared(count_of_im,category,image,buffer) private(j,i)
    {
#pragma omp single
        cout <<__FUNCTION__ << ": Number of threads: "<< omp_get_num_threads() << endl;


#pragma omp for
        for (i = 0; i < count_of_im; i++) {
            for (j = 0; j < 3073; j++) {

                if (j == 0) {
                    category[i][1] = i;
                    category[i][0] = static_cast<int>(static_cast<bitset<8>>(buffer[i * 3073]).to_ulong());

                } else {
                    image[i][j - 1] = static_cast<int>(static_cast<bitset<8>>(buffer[i * 3073 + j]).to_ulong());
                }

            }

        }
    }
    cout<<__FUNCTION__ <<": Done..."<<endl;
    delete[] buffer;
    delete file;
}

void normalization( double ** category, double ** image, int count_of_image, int num_threads) {
    cout<<__FUNCTION__ <<": Normalization Gauss"<<endl;
    int i{0}, j{0};
    double mean{0};
    double st{0};
    mean = get_mean(image, count_of_image);
    st = get_std(mean, count_of_image, image);

    omp_set_dynamic(0);
    omp_set_num_threads(num_threads);
#pragma omp parallel default(shared) private(i, j)
    {
#pragma omp single
        cout <<__FUNCTION__ << ": Number of threads: "<< omp_get_num_threads() << endl;

#pragma omp for
        for (i = 0; i < count_of_image; ++i)
            for (j = 0; j < nbits; ++j)
                image[i][j] = (image[i][j] - mean) / st;
    }


}

void save(double ** category, double ** image, int count_of_image, char * path_to_file)
{
    fstream file;
    file.open(path_to_file,ios_base::out);

    for (int i = 0; i < count_of_image; ++i) {
        file<<category[i][1]<<"\t"<<category[i][0]<<"\t";
        for (int j = 0; j < 10; ++j) {
            file<<image[i][j]<<"\t";
        }
        file<<endl;
    }
    file.close();




}

int main(int argc, char **argv) {

    int num_threads=atoi(argv[1]);
    double ** category;
    double ** image;
    category = new double* [count_of_data];
    image=new double * [count_of_data];

    for (int k = 0; k < count_of_data; ++k) {
        category[k]=new double [2];
        image[k]=new double[nbits];
    }

    get_images(argv[2],count_of_data, category, image, num_threads);
    normalization(category,image,count_of_data, num_threads);
    save(category,image,count_of_data,"data_gauss.txt");

    for (int l = 0; l < count_of_data; ++l) {
        delete[] category[l] ;
        delete[] image[l];
    }

    delete [] category;
    delete [] image;



    return 0;
}
