#include <iostream>
#include <fstream>
#include <bitset>
#include <string>
#include <omp.h>
#include <cmath>

constexpr int nbits=3072;
constexpr int file_size = 30730000;

using namespace std;

typedef void(* distance_point)(double **, double **, double *, int) ;

double get_mean(double** image,int count_of_image)
{
    int sum{0};
    int i{0}, j{0};
#pragma omp parallel default(shared) private(i, j)
    {

#pragma omp for schedule(static) reduction(+:sum)
        for (i = 0; i < count_of_image; ++i) {
            for (j = 0; j < nbits; ++j) {
                sum += static_cast<int>(image[i][j]) ;
            }
        }
    }

    return static_cast<double>(sum)/ static_cast<double>(nbits*count_of_image);
}

double get_std(double mean, int count_of_im, double ** image)
{
    double st=0;
    for (int i = 0; i < count_of_im; ++i) {
        for (int j = 0; j < nbits; ++j) {
            st+=(image[i][j]-mean)*(image[i][j]-mean);
        }
    }
    return sqrt(st);
}

int * min_max(double ** image, int count_of_image, int num_threads)
{

    int i{0}, j{0};
    int min{255},max{0};
    int * tab = new int[2];





        for (i = 0; i < count_of_image; ++i) {
            for (j = 0; j < 3072; ++j) {
                if (min > image[i][j])
                    min = image[i][j];
                else if (max < image[i][j])
                    max = image[i][j];
            }
        }


    tab[0]=min;
    tab[1]=max;
    return tab;
}

void get_images(string name, int count_of_im, double ** category, double** image, int num_threads)
{

    cout<<__FUNCTION__ <<" Loading..."<<endl;
    ifstream *file = new ifstream(name.c_str(), ios_base::binary | ios_base::in);

    char * buffer = new char[file_size];

    file->seekg(0, std::ios::beg);
    file->read(buffer,file_size);
    file->close();
    cout<<__FUNCTION__ <<" File size: "<<file_size<<endl;

    int i=0, j=0;



cout<<__FUNCTION__ <<" Preaparing..."<<endl;
omp_set_dynamic(0);
omp_set_num_threads(num_threads);
#pragma omp parallel default(shared) private(j,i)
    {
#pragma omp single
            cout <<__FUNCTION__ << "Number of threads: "<< omp_get_num_threads() << endl;


#pragma omp for schedule(static, 25)
        for (i = 0; i < count_of_im; i++) {
            for (j = 0; j < 3073; j++) {

                if (j == 0) {
                    category[i][1] = i;
                    category[i][2] = 0;
                    category[i][0] = static_cast<int>(static_cast<bitset<8>>(buffer[i * 3073]).to_ulong());

                } else {
                    image[i][j - 1] = static_cast<int>(static_cast<bitset<8>>(buffer[i * 3073 + j]).to_ulong());
                }

            }

        }
    }
    cout<<__FUNCTION__ <<" Done..."<<endl;
    delete[] buffer;
    delete file;
}

void normalization( double ** category, double ** image, int count_of_image, char which_one, int num_threads) {

    cout<<__FUNCTION__ <<" Normalization ..."<<endl;

    switch (which_one) {
        case 'm': {
            cout<<__FUNCTION__ <<" Normalization min-max"<<endl;
            int *mi_ma;
            int i{0}, j{0};
            int min{0}, max{0};

            mi_ma = min_max(image, count_of_image, num_threads);

            min = mi_ma[0];
            max = mi_ma[1];
            cout<<min<<" aa "<<max<<endl;
            omp_set_dynamic(0);
            omp_set_num_threads(num_threads);
   // #pragma omp parallel default(shared) shared(image, count_of_image, min, max) private(i, j)
            {



    //#pragma omp for schedule(static)
                for (i = 0; i < count_of_image; ++i)
                    for (j = 0; j < nbits; ++j)
                        image[i][j] = (image[i][j] - min) / (max - min);
            }
            //delete [] mi_ma;
            break;
        }
        case 'g': {
            cout<<__FUNCTION__ <<" Normalization Gauss"<<endl;
            int i{0}, j{0};
            double mean{0};
            double st{0};
            mean = get_mean(image, count_of_image);
            st = get_std(mean, count_of_image, image);

            omp_set_dynamic(0);
            omp_set_num_threads(num_threads);
    #pragma omp parallel default(shared)  private(i, j)
            {
    #pragma omp single
                    cout <<__FUNCTION__ << "Number of threads: "<< omp_get_num_threads() << endl;

    #pragma omp for schedule(static)
                for (i = 0; i < count_of_image; ++i)
                    for (j = 0; j < nbits; ++j)
                        image[i][j] = (image[i][j] - mean) / st;
            }
            break;
        }
        default:
            cout<<__FUNCTION__ <<"There is nothing to do!"<<endl;
            return;

    }
    cout<<__FUNCTION__ <<" Done..."<<endl;
    return;
}

void distance_euclid(double ** image_training, double ** category_training, double * image_testing, int count_of_training_im)
{

    double x{0.0};
    for (int i = 0; i <count_of_training_im ; ++i) {

        for (int j = 0; j < nbits ; ++j) {
            x+=(image_testing[j]-image_training[i][j])*(image_testing[j]-image_training[i][j]);
        }

        for (int k = 0; k < count_of_training_im; ++k) {
            if(category_training[k][1]==i) {
                category_training[k][2] = sqrt(x);
                break;
            }
        }
        x=0.0;

    }

}

void distance_manhattan(double ** image_training, double ** category_training, double * image_testing, int count_of_training_im)
{

    double x{0.0};
    for (int i = 0; i <count_of_training_im ; ++i) {

        for (int j = 0; j < nbits ; ++j) {
            x+=fabs(image_testing[j]-image_training[i][j]);
        }

        for (int k = 0; k < count_of_training_im; ++k) {
            if(category_training[k][1]==i) {
                category_training[k][2] = x;
                break;
            }
        }
        x=0.0;

    }

}

void distance_chessboard(double ** image_training, double ** category_training, double * image_testing, int count_of_training_im)
{

    double x{0.0};
    double tmp{x};


    for (int i = 0; i <count_of_training_im ; ++i) {

        for (int j = 0; j < nbits ; ++j) {
            x=fabs(image_testing[j]-image_training[i][j]);
            x>tmp ? tmp=x : x ;
        }

        for (int k = 0; k < count_of_training_im; ++k) {
            if(category_training[k][1]==i) {
                category_training[k][2] = x;
                break;
            }
        }
        x=0.0;

    }

}

int partition(double **category, int p, int r)
{

    double x[3], w[3];
    double * t;
    int i{p}, j{r};
    x[0] = category[p][0];
    x[1] = category[p][1];
    x[2] = category[p][2];

    while(true)
    {
        while(category[j][2]>x[2])
            j--;
        while(category[i][2]<x[2])
            i++;

        if(i<j)
        {
            t=category[i];
            category[i]= category[j];
            category[j]=t;
            j--;
            i++;

        } else
            return j;
    }
}

void sort(double ** category_training, int first, int count_of_tested_m_one)
{

    int p=first,r=count_of_tested_m_one;
    int l;
    if(p<r)
    {
        l = partition(category_training,p,r);
        sort(category_training,p,l);
        sort(category_training,l+1,r);
    }


}


double knn_majority(double** image_training, double ** image_testing, double ** category_training, double ** category_test, int count_of_training_im, int count_of_tested_im, int num_threads, distance_point distance)
{
    cout<<__FUNCTION__ <<" Knn, start!"<<endl;
int i{0}, j{0}, k{0}, counter{0};
double neighborhood[10]={0,0,0,0,0,0,0,0,0,0};
double tmp{0.0};
double **c_tmp{nullptr};
double start{0}, end{0};
omp_set_dynamic(0);
omp_set_num_threads(num_threads);


#pragma omp parallel default(shared)  private(i,j,k,neighborhood,c_tmp,tmp)
    {
#pragma omp single
        cout <<__FUNCTION__ << "Number of threads: "<< omp_get_num_threads() << endl;

        c_tmp = new double *[count_of_training_im];
        for (int l = 0; l < count_of_training_im; ++l) {
            c_tmp[l] = new double[3];
            c_tmp[l][0] = category_training[l][0];
            c_tmp[l][1] = category_training[l][1];
            c_tmp[l][2] = category_training[l][2];
        }

#pragma omp for schedule(static)
        for (i = 0; i < count_of_tested_im; ++i) {

            distance(image_training, c_tmp, image_testing[i], count_of_training_im);
            sort(c_tmp, 0, count_of_training_im - 1);

            for (k = 0; k < 5; ++k) {
                neighborhood[static_cast<int>(c_tmp[k][0])]+=1./static_cast<int>(c_tmp[k][2]);
            }

            for (int l = 0; l < 10; ++l) {
                if (neighborhood[l] > tmp) {
                    category_test[i][2] = l;
                    tmp = neighborhood[l];
                }

            }
            tmp = 0;

            for (int m = 0; m < 10; ++m) {
                neighborhood[m]=0;
            }

        }

        for (int n = 0; n < count_of_training_im; ++n) {
            delete [] c_tmp[n];
        }
        delete [] c_tmp;
    }


    for (int n = 0; n < count_of_tested_im; ++n) {
        if(category_test[n][2]==category_test[n][0])
            counter++;
    }

    cout<<__FUNCTION__ <<" Done!"<<endl;
    ;
    return static_cast<double>(counter)/static_cast<double>(count_of_tested_im);
}

int main(int argc, char **argv) {

    int num_threads=atoi(argv[1]);
    int count_of_training=10000;
    int count_of_tested=1500;
    double ** category;
    double ** image;
    double ** category_t;
    double ** image_t;


    category_t = new double *[count_of_training];
    image_t = new double * [count_of_training];
    category = new double* [count_of_tested];
    image=new double * [count_of_tested];


    for (int k = 0; k < count_of_tested; ++k) {
        category[k]=new double [3];
        image[k]=new double[nbits];
    }
    for (int j = 0; j < count_of_training ; ++j) {
        category_t[j] = new double [3];
        image_t[j] = new double [nbits];

    }

    get_images(argv[2],count_of_tested, category, image, num_threads);
    get_images(argv[3], count_of_training, category_t, image_t, num_threads);



    normalization(category_t,image_t,count_of_training,*argv[4], num_threads);
    normalization(category,image,count_of_tested,*argv[4], num_threads);



    cout<< knn_majority(image_t,image,category_t,category,count_of_training,count_of_tested, 4, distance_euclid)<<endl;






    for (int k = 0; k < count_of_training; ++k) {
        delete[] category_t[k];
        delete[] image_t[k];
    }

    for (int l = 0; l < count_of_tested; ++l) {
        delete[] category[l] ;
        delete[] image[l];
    }

    delete [] category_t;
    delete [] category;
    delete [] image_t;
    delete [] image;
   return 0;
}
