#include <iostream>
#include <omp.h>
#include <fstream>
#include <bitset>

constexpr int nbits=3072;
constexpr int file_size = 30730000;
constexpr int count_of_data=500;
using namespace std;

int * min_max(double ** image, int count_of_image, int num_threads)
{

    int i{0}, j{0};
    int min{255},max{0};
    int * tab = new int[2];

    omp_set_dynamic(0);
    omp_set_num_threads(num_threads);
#pragma omp parallel default(shared) shared(image,count_of_image) private(i, j, min,max)
    {

#pragma omp single
        cout << __FUNCTION__ <<" Number of threads: "<< omp_get_num_threads() << endl;


#pragma omp for schedule(static)
        for (i = 0; i < count_of_image; ++i) {
            for (j = 0; j < 3072; ++j) {
                if (min > image[i][j])
                    min = image[i][j];
                else if (max < image[i][j])
                    max = image[i][j];
            }
        }
    }

    tab[0]=min;
    tab[1]=max;
    return tab;
}

void get_images(string name, int count_of_im, double ** category, double** image, int num_threads)
{

    cout<<__FUNCTION__ <<" Loading..."<<endl;
    ifstream *file = new ifstream(name.c_str(), ios_base::binary | ios_base::in);

    char * buffer = new char[file_size];

    file->seekg(0, std::ios::beg);
    file->read(buffer,file_size);
    file->close();
    cout<<__FUNCTION__ <<" File size: "<<file_size<<endl;

    int i=0, j=0;



    cout<<__FUNCTION__ <<" Preaparing..."<<endl;
    omp_set_dynamic(0);
    omp_set_num_threads(num_threads);
#pragma omp parallel default(shared) shared(count_of_im,category,image,buffer) private(j,i)
    {
#pragma omp single
        cout <<__FUNCTION__ << " Number of threads: "<< omp_get_num_threads() << endl;


#pragma omp for schedule(static, 25)
        for (i = 0; i < count_of_im; i++) {
            for (j = 0; j < 3073; j++) {

                if (j == 0) {
                    category[i][1] = i;
                    category[i][0] = static_cast<int>(static_cast<bitset<8>>(buffer[i * 3073]).to_ulong());

                } else {
                    image[i][j - 1] = static_cast<int>(static_cast<bitset<8>>(buffer[i * 3073 + j]).to_ulong());
                }

            }

        }
    }
    cout<<__FUNCTION__ <<" Done..."<<endl;
    delete[] buffer;
    delete file;
}

void normalization( double ** category, double ** image, int count_of_image, int num_threads) {
            cout<<__FUNCTION__ <<" Normalization min-max"<<endl;
            int *mi_ma;
            int i{0}, j{0};
            int min{0}, max{0};

            mi_ma = min_max(image, count_of_image, num_threads);

            min = mi_ma[0];
            max = mi_ma[1];

            omp_set_dynamic(0);
            omp_set_num_threads(num_threads);
#pragma omp parallel default(shared) shared(image, count_of_image, min, max) private(i, j)
            {

#pragma omp single
                cout <<__FUNCTION__ << ": Number of threads: "<< omp_get_num_threads() << endl;

#pragma omp for schedule(static)
                for (i = 0; i < count_of_image; ++i)
                    for (j = 0; j < nbits; ++j)
                        image[i][j] = (image[i][j] - min) / (max - min);
            }

}

void save(double ** category, double ** image, int count_of_image, char * path_to_file)
{
    fstream file;
    file.open(path_to_file,ios_base::out);

    for (int i = 0; i < count_of_image; ++i) {
        file<<category[i][1]<<"\t"<<category[i][0]<<"\t";
        for (int j = 0; j < nbits; ++j) {
            file<<image[i][j]<<"\t";
        }
        file<<endl;
    }
    file.close();




}

int main(int argc, char **argv) {

    int num_threads=atoi(argv[1]);
    double ** category;
    double ** image;
    category = new double* [count_of_data];
    image=new double * [count_of_data];

    for (int k = 0; k < count_of_data; ++k) {
        category[k]=new double [2];
        image[k]=new double[nbits];
    }

    get_images(argv[2],count_of_data, category, image, num_threads);
    normalization(category,image,count_of_data, num_threads);
    save(category,image,count_of_data,"data_min_max.txt");

    for (int l = 0; l < count_of_data; ++l) {
        delete[] category[l] ;
        delete[] image[l];
    }

    delete [] category;
    delete [] image;



    return 0;
}
